/*
  Copyright (C) 2020 Alessandro Languasco & Luca Righi
*/
#include "compute.h"

#define REAL 0
#define IMAG 1

long double Eulergamma = 0.57721566490153286060651209008240243104;
long double pi = 3.1415926535897932384626433832795028841971693993751;

/* for odd/even results */
struct result {
  long double res;
  long double max;
  long double min;
  long double lambda;
  long double Kummer;
  long double maxL;
  long double minL;
  long double maxLprime;
  long double minLprime;
  
};

struct measure {
  long double norm2;
  long double bnorm2;
  long double bnorm2even;
  long double bnorm2odd;
  long double norm1;
  long double bnorm1;
  long double bnorm1even;
  long double bnorm1odd;
  long double norm00;
  long double bnorm00;
  long double bnorm00even;
  long double bnorm00odd;
  long double prec2;
  long double prec1;
  long double prec00;
};

struct norms {
  struct measure norm_ak;
  struct measure norm_lgodd;
  struct measure norm_lgeven;
  struct measure norm_s;
};

/* mmap.c */
void * mapRO_File2Array(const char *filename, int* fd, size_t arraySize, size_t typeSize);
long double *mapRWCreate_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize);
long double *mapRW_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize);
void unmap_File2Array(void * mmappedData, size_t filesize, int fd);
char * mallocFilename(const char * basename, const char* suffix);
void * mkmmapcopy(void * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix, const char mode, int *fd);
fftwl_complex * mapfile(const char * name, const char * basename, long int size, int * fd);
void  print_time(const char * message, double startplan, double endplan);
void * coniugateAndDivideFFTFileInMappedFile(fftwl_complex * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix);
void * copyAndNormalizeFFTFileInMappedFile(fftwl_complex * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix);
void printcomplexvector(fftwl_complex * v, ptrdiff_t start, ptrdiff_t step, ptrdiff_t size, const char * name, FILE *stream);
/* s.c */
long int lpow_mod(long int base, long int exp, long int mod);
long double s(long double aq, int precision, int verbose);

struct result readResult(const char * basename, const char * name) {
  const char * funcname = "readResult()";
  
  struct result * r = NULL;
  if ((r = malloc(sizeof(struct result))) == NULL) {
    fprintf(stderr, "%s: unable to allocate result. Giving up\n", funcname);
    exit(1);
  }
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "r")) == NULL) {
    fprintf(stderr, "%s: can't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }
  
  if (fscanf(f, "%Lf\n", &(r->res)) != 1) {
    fprintf(stderr, "%s: unable to read 'res' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf\n", &(r->max)) != 1) {
    fprintf(stderr, "%s: unable to read 'max' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf", &(r->min)) != 1) {
    fprintf(stderr, "%s: unable to read 'min' from %s: %s\n", funcname, filename, strerror(errno));
  }

  /* */
  if (fscanf(f, "%Lf", &(r->lambda)) != 1) {
    fprintf(stderr, "%s: unable to read 'lambda' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf", &(r->Kummer)) != 1) {
    fprintf(stderr, "%s: unable to read 'Kummer' from %s: %s\n", funcname, filename, strerror(errno));
  }

  if (fscanf(f, "%Lf\n", &(r->maxL)) != 1) {
    fprintf(stderr, "%s: unable to read 'maxL' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf", &(r->minL)) != 1) {
    fprintf(stderr, "%s: unable to read 'minL' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf\n", &(r->maxLprime)) != 1) {
    fprintf(stderr, "%s: unable to read 'maxLprime' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if (fscanf(f, "%Lf", &(r->minLprime)) != 1) {
    fprintf(stderr, "%s: unable to read 'minLprime' from %s: %s\n", funcname, filename, strerror(errno));
  }
  /* */
     

  fclose(f);

  // printf("%s: readed %Lf, %Lf, %Lf from %s\n", funcname, r->res, r->max, r->min, filename);
  printf("%s: read %Lf, %Lf, %Lf, %Lf, %Lf, %Lf, %Lf, %Lf, %Lf from %s\n", funcname, 
	 r->res, r->max, r->min, r->lambda, r->Kummer, r->maxL, r->minL,
	 r->maxLprime, r->minLprime, filename);
  
  return *r;
}

void writeResult(struct result r, const char * basename, const char * name) {

  const char * funcname = "writeResult()";
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "w")) == NULL) {
    fprintf(stderr, "%s: san't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }

  fprintf(f, "%Lf\n", r.res);
  fprintf(f, "%Lf\n", r.max);
  fprintf(f, "%Lf\n", r.min);
  fprintf(f, "%Lf\n", r.lambda);
  fprintf(f, "%Lf\n", r.Kummer);
  fprintf(f, "%Lf\n", r.maxL);
  fprintf(f, "%Lf\n", r.minL);
  fprintf(f, "%Lf\n", r.maxLprime);
  fprintf(f, "%Lf\n", r.minLprime);

  fclose(f);

  return;
}

fftwl_complex * mkfftin(ptrdiff_t nhalf, const char *basename, int *fd_fftin) {

  char * funcname = "mkfftin()";
  char * fftinFilename = mallocFilename(basename, N_FFTINPUT);

  fftwl_complex * fftin = (fftwl_complex *) mapRWCreate_File2Array(fftinFilename, fd_fftin, nhalf, sizeof(fftwl_complex));
  
  if (fftin == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftinFilename);
    return NULL;
  }

  return fftin;
}

fftwl_complex * mkfftout(ptrdiff_t nhalf, const char *basename, int *fd_fftout) {

  char * funcname = "mkfftout()";
  char * fftoutFilename = mallocFilename(basename, N_FFTOUTPUT);

  fftwl_complex * fftout = (fftwl_complex *) mapRWCreate_File2Array(fftoutFilename, fd_fftout, nhalf, sizeof(fftwl_complex));
  if (fftout == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftoutFilename);
    return 0;
  }

  return fftout;
}

fftwl_complex * mkfftinverse(ptrdiff_t nhalf, const char *basename, int *fd_fftinverse) {

  char * funcname = "mkfftinverse()";
  char * fftinverseFilename = mallocFilename(basename, N_FFTINVERSE);

  fftwl_complex * fftinverse = (fftwl_complex *) mapRWCreate_File2Array(fftinverseFilename, fd_fftinverse, nhalf, sizeof(fftwl_complex));
  if (fftinverse == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftinverseFilename);
    return 0;
  }

  return fftinverse;
}


fftwl_plan mkplan(fftwl_complex *fftin, fftwl_complex *fftout, ptrdiff_t nhalf) {

  if (fftin == fftout) {
    /* same array. Wrong. This is an outplace FFT */
    fprintf(stderr,  "*************************************************\n");
    fprintf(stderr,  "*** ERROR: fftin == fftout. FFT in not INPLACE. *\n");
    fprintf(stderr,  "*************************************************\n");
    return NULL;
  }
  
  clock_t startplan = clock();
  
  if (fftwl_import_wisdom_from_filename("wisdom-outplace-fftwl-estimate.txt")) {
    fprintf(stderr,  "*** Reading wisdom-outplace-fftwl-estimate \n");
  } else {
    fprintf(stderr,  "*** Generating wisdom-outplace-fftwl-estimate \n");
  }

  int rank=1;
  fftw_iodim64 * dims = malloc(rank * sizeof(fftw_iodim64));
  dims[0].n= nhalf;   /* length of the trasform */
  dims[0].is=1;
  dims[0].os=1; 
  int howmany_rank=1; 
  fftw_iodim64 * howmany_dims = malloc(howmany_rank * sizeof(fftw_iodim64)); 
  howmany_dims[0].n=1;
  howmany_dims[0].is=1;
  howmany_dims[0].os=1;
  
  fftwl_plan plan = fftwl_plan_guru64_dft(rank, /*rank=1 means monodimensional vector*/
              dims, /* nhalf */
              howmany_rank,
              howmany_dims,
              fftin,   /*input for OUTPLACE transforms*/ 
              fftout,  /*output for OUTPLACE transforms*/ 
              FFTW_FORWARD, /* -1 because it is the sign of FORWARD transform */
              FFTW_ESTIMATE); /* flags */
              
  if (plan == NULL) { 
    fprintf(stderr,"plan creation failed\n");
    return 0;
  }
    
  /*  saving  a  generated  wisdom  for  future  runs  */
  if (!fftwl_import_wisdom_from_filename("wisdom-outplace-fftwl-estimate.txt")) {
    fprintf(stderr,  "***  Saving wisdom-outplace-fftwl-estimate.txt \n");
    fftwl_export_wisdom_to_filename("wisdom-outplace-fftwl-estimate.txt"); 
  }
  
  clock_t endplan = clock(); 

  print_time("Plan generation", startplan, endplan);

  return plan;
}

fftwl_plan mkinverseplan(fftwl_complex *fftin, fftwl_complex *fftout, ptrdiff_t nhalf) {

  if (fftin == fftout) {
    /* same array. Wrong. This is an outplace FFT */
    fprintf(stderr,  "*************************************************\n");
    fprintf(stderr,  "*** ERROR: fftin == fftout. FFT in not INPLACE. *\n");
    fprintf(stderr,  "*************************************************\n");
    return NULL;
  }
  
  clock_t startplan = clock();
  
  if (fftwl_import_wisdom_from_filename("wisdom-outplace-inverse-fftwl-estimate.txt")) {
    fprintf(stderr,  "*** Reading wisdom-outplace-inverse-fftwl-estimate \n");
  } else {
    fprintf(stderr,  "*** Generating wisdom-outplace-inverse-fftwl-estimate \n");
  }

  int rank=1;
  fftw_iodim64 * dims = malloc(rank * sizeof(fftw_iodim64));
  dims[0].n= nhalf;   /* length of the trasform */
  dims[0].is=1;
  dims[0].os=1; 
  int howmany_rank=1; 
  fftw_iodim64 * howmany_dims = malloc(howmany_rank * sizeof(fftw_iodim64)); 
  howmany_dims[0].n=1;
  howmany_dims[0].is=1;
  howmany_dims[0].os=1;
  
  fftwl_plan plan = fftwl_plan_guru64_dft(rank, /*rank=1 means monodimensional vector*/
              dims, /* nhalf */
              howmany_rank,
              howmany_dims,
              fftin,   /*input for OUTPLACE transforms*/ 
              fftout,  /*output for OUTPLACE transforms*/ 
              FFTW_BACKWARD, /* 1 because it is the sign of BACKWARD transform */
              FFTW_ESTIMATE); /* flags */
              
  if (plan == NULL) { 
    fprintf(stderr,"inverse plan creation failed\n");
    return 0;
  }
    
  /*  saving  a  generated  wisdom  for  future  runs  */
  if (!fftwl_import_wisdom_from_filename("wisdom-outplace-inverse-fftwl-estimate.txt")) {
    fprintf(stderr,  "***  Saving wisdom-outplace-inverse-fftwl-estimate.txt \n");
    fftwl_export_wisdom_to_filename("wisdom-outplace-inverse-fftwl-estimate.txt"); 
  }
  
  clock_t endplan = clock(); 

  print_time("Inverse plan generation time", startplan, endplan);

  return plan;
}

fftwl_complex * mkLGodd(fftwl_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_LGodd) {

  long double pi = 3.1415926535897932384626433832795028841971693993751;
  long double sinvalue;
  long double cosvalue;
  long double logpi = logl(pi);
  long double twist = 0.0L;
  long double angle = 0.0L; 
  long double aux;
  long double d;
  long int a;
  ptrdiff_t k;
  
  /* 
   * STEP: mkLGodd
   *  need: fftin mmapped array
   *  provide: file <basename>-LGodd.ld
   */
  
  clock_t startclock = clock();  /* start clock */

  angle = (long double) pi / nhalf;
  twist = 0.0L; 
  a = 1; 

  for (k = 0; k <nhalf ; k++){   
     cosvalue = cosl(twist);
     sinvalue = sinl(twist);  
     d = (long double) a/q ;
     aux = (2.0) * lgammal( d ) + logl( sinl(pi * d)) - logpi ;
     fftin[k][REAL] = aux * cosvalue;  
     fftin[k][IMAG] = (-1.0L) * aux * sinvalue; 
     a = (a * primroot) % q;  
     twist = twist + angle;
  }

  clock_t endclock = clock();  /* end clock */  
  print_time("LNGammaOdd computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwl_complex * lgodd = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_LGODD, 'r', fd_LGodd);

  return lgodd;

}

fftwl_complex * fftLGodd(fftwl_complex *fftin, fftwl_complex *fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_LGoddfft) {
  
  /* 
   * STEP: fftLGodd
   *  need: fftin mmapped array with LGodd data
   *  provide: fft of fftin in array fftout
   *  provide: copy of fftout in file <basename>-LGoddfft.ld if basename != NULL
   */

  clock_t startfft = clock(); /* start clock */
  fftwl_execute(plan); 
  clock_t endfft = clock(); /* stop clock */
  print_time("FFT(LGodd)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwl_complex * lgoddfft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTLGODD, 'r', fd_LGoddfft);

  return lgoddfft;

}

/* to be developed */
fftwl_complex * fft_INV(fftwl_complex *fftin, fftwl_complex *fftout,
			fftwl_plan iplan, long int nhalf,
			const char *basename,
			const char *fftname,
			const char *fftinvname,
			int *fd_INV) {
  
  /* 
   * STEP: fft_INV
   *  need: fftin mmapped array to be filled with fft data (from fftname)
   *  need: basename: basename for the files
   *  need: iplan: plan for the inverse FFT
   *  need: fftname: suffix for basename to open th file with the input fft data
   *  need: fftinvname: suffix for basename to write the file with the fft inverse data
   *  provide: fft(fftin) in array fftout 
  *  provide: copy of fftout in file <basename>-<fftinvname>
   */

  char* funcName = "FFT_INV()";

  clock_t start_bcopy = clock(); /* start clock */
  if ((fftin = copyAndNormalizeFFTFileInMappedFile(fftin, nhalf, sizeof(fftwl_complex), basename, fftname)) == NULL) {
    fprintf(stderr, "%s: *** ERROR: unable to loadFilenameInMappedFile(%s,%s)\n", funcName, basename, fftname);
    exit(0);    
  }
  clock_t end_bcopy = clock(); /* stop clock */
  print_time("FFT_INV(): Initial copy and normalization time\n", start_bcopy, end_bcopy);
  
  clock_t startfft_INV = clock(); /* start clock */
  fftwl_execute(iplan); 
  clock_t endfft_INV = clock(); /* stop clock */
  print_time("FFT_INV(): fftwl inverse plan execute time", startfft_INV, endfft_INV);
  fflush(stdout);

  clock_t start_copy = clock(); /* start clock */
  /* make the copy of the output in basename-fftinvname */
  mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, fftinvname, '-', fd_INV);
  clock_t end_copy = clock(); /* stop clock */
  print_time("FFT_INV(): Result copy", start_copy, end_copy);
  
  return fftout;

}

/*
 * compute Norm00, Norm1 and Norm2 for |base - inverse| / |base|
 */
int computenorms(const char *basename, const char *base, const char *inverse, long int size, struct measure *m, int evenodd) {
  const char* funcname = "norm2diff()";
  fftwl_complex * b = NULL;
  fftwl_complex * i = NULL;
  long double x, y;
  int fd_b, fd_i;

  if ((b = mapfile(base, basename, size, &fd_b)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, base);
    return 0;
  }

  if ((i = mapfile(inverse, basename, size, &fd_i)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, inverse);
    return 0;
  }

  /* NORM2 sum vars for ||b - i|| */
  long double SUM2 = 0.0;

  /* NORM2 sum vars for ||b|| (the original vector) */
  long double bSUM2 = 0.0;

  /* NORM1 sum vars for ||b - i|| */
  long double SUM1 = 0.0;
  
  /* NORM1 sum vars for ||b|| (the original vector) */
  long double bSUM1 = 0.0;
  
  /* NORM00 vars for ||b - i|| */
  long double DMAX = 0.0;
  long double DTEMP = 0.0;

  /* NORM00 vars for ||b|| */
  long double BMAX = 0.0L;
  long double BTEMP = 0.0L;

  for (long int k = 0; k < size; k++) {
    /* difference vector component */
    x = b[k][REAL] - i[k][REAL];
    y = b[k][IMAG] - i[k][IMAG];

    printf("%LE + %LEi, ", x, y);

    /* NORM2 b-i */
    SUM2 += (x * x + y * y);

    /* NORM2 b */
    bSUM2 += (b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);

    /* NORM1 b-i */
    SUM1 += sqrtl(x * x + y * y);

    /* NORM1 b */
    bSUM1 += sqrtl(b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);

    /* NORM00 b - i */
    DTEMP = sqrtl(x * x + y * y);
    if (DTEMP > DMAX) {
      DMAX = DTEMP;
      // DPOS = k;
    }

    /* NORM00 b */
    BTEMP = sqrtl(b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);
    if (BTEMP > BMAX) {
      BMAX = BTEMP;
      // BPOS = k;
    }
  }

  printf("]\n");
  
  m->norm2 = sqrtl(SUM2);
  m->bnorm2 = sqrtl(bSUM2);

  m->norm1 = SUM1;
  m->bnorm1 = bSUM1;

  // printf("DEBUG: norm2 = %32.30LE, norm1 = %32.30LE\n", m->norm2, m->norm1);
  
  m->norm00 = DMAX;
  m->bnorm00 = BMAX;

  if (fabsl(m->bnorm2) > 0.000001) {
    m->prec2 = m->norm2 / m->bnorm2;
  } else {
    m->prec2 = 0.0;
  }

  if (fabsl(m->bnorm1) > 0.000001) {
    m->prec1 = m->norm1 / m->bnorm1;
  } else {
    m->prec1 = 0.0;
  }

  if (fabsl(m->bnorm00) > 0.000001) {
    m->prec00 = m->norm00 / m->bnorm00;
  } else {
    m->prec00 = 0.0;
  }
  
  unmap_File2Array(b, size, fd_b);
  unmap_File2Array(i, size, fd_i);

  return 1;
}

void nullmeasure(struct measure *m) {
  if (m == NULL) {
    fprintf(stderr, "nullmeasure(): measure == NULL\n");
    return;
  }

  m->norm2 = 0.0;
  m->bnorm2 = 0.0;
  m->bnorm2even = 0.0;
  m->bnorm2odd = 0.0;
  m->norm1 = 0.0;
  m->bnorm1 = 0.0;
  m->bnorm1even = 0.0;
  m->bnorm1odd = 0.0;
  m->norm00 = 0.0;
  m->bnorm00 = 0.0;
  m->bnorm00even = 0.0;
  m->bnorm00odd = 0.0;
  m->prec2 = 0.0;
  m->prec1 = 0.0;
  m->prec00 = 0.0;

}

/*
 * compute Norm00, Norm1 and Norm2 for |base - inverse| / |base|
 */
int computenormskahan(const char *basename, const char *base, const char *inverse, long int size, struct measure *m, int evenodd) {
  const char* funcname = "norm2diff()";
  fftwl_complex * b = NULL;
  fftwl_complex * i = NULL;
  long double x, y, belem;
  int fd_b, fd_i;

  int DEBUG=0;

  if ((b = mapfile(base, basename, size, &fd_b)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, base);
    nullmeasure(m);
    return 0;
  }

  if ((i = mapfile(inverse, basename, size, &fd_i)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, inverse);
    nullmeasure(m);
    return 0;
  }

  /* NORM2 Kahan sum vars for ||b - i|| */
  long double SUM2 = 0.0;
  long double C2 = 0.0;
  long double Y2 = 0.0;
  long double T2 = 0.0;

  /* NORM2 Kahan sum vars for ||b|| (the original vector) */
  long double bSUM2 = 0.0;
  long double bC2 = 0.0;
  long double bY2 = 0.0;
  long double bT2 = 0.0;

  /* NORM2 Kahan sum vars for ||b|| (the original vector) EVEN */
  long double bSUM2e = 0.0;
  long double bC2e = 0.0;
  long double bY2e = 0.0;
  long double bT2e = 0.0;

  /* NORM2 Kahan sum vars for ||b|| (the original vector) ODD */
  long double bSUM2o = 0.0;
  long double bC2o = 0.0;
  long double bY2o = 0.0;
  long double bT2o = 0.0;

  /* NORM1 Kahan sum vars for ||b - i|| */
  long double SUM1 = 0.0;
  long double C1 = 0.0;
  long double Y1 = 0.0;
  long double T1 = 0.0;
  
  /* NORM1 Kahan sum vars for ||b|| (the original vector) */
  long double bSUM1 = 0.0;
  long double bC1 = 0.0;
  long double bY1 = 0.0;
  long double bT1 = 0.0;

  /* NORM1 Kahan sum vars for ||b|| (the original vector) EVEN */
  long double bSUM1e = 0.0;
  long double bC1e = 0.0;
  long double bY1e = 0.0;
  long double bT1e = 0.0;

  /* NORM1 Kahan sum vars for ||b|| (the original vector) ODD */
  long double bSUM1o = 0.0;
  long double bC1o = 0.0;
  long double bY1o = 0.0;
  long double bT1o = 0.0;
  
  /* NORM00 vars for ||b - i|| */
  long double DMAX = 0.0;
  long double DTEMP = 0.0;

  /* NORM00 vars for ||b|| */
  long double BMAX = 0.0L;
  long double BTEMP = 0.0L;

  /* NORM00 vars for ||b|| EVEN */
  long double BMAXe = 0.0L;

  /* NORM00 vars for ||b|| ODD */
  long double BMAXo = 0.0L;

  if (DEBUG) {
    printcomplexvector(b, 0, 1, size, base, stdout); // complete vector
    printcomplexvector(b, 0, 2, size, base, stdout); // even components
    printcomplexvector(b, 1, 2, size, base, stdout); // odd component
  }
  
  for (long int k = 0; k < size; k++) {

    /* difference vector component */
    x = b[k][REAL] - i[k][REAL];
    y = b[k][IMAG] - i[k][IMAG];

    /* NORM2 b-i */
    Y2 = (x * x + y * y) - C2;   /* qui sommo i quadrati dei valori assoluti delle componenti poi dovro` fare la radice quadrata */
    T2 = SUM2 + Y2;
    C2 = (T2 - SUM2) - Y2;
    SUM2 = T2;

    /* NORM2 b */
    belem = (b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);
    bY2 = belem - bC2;
    bT2 = bSUM2 + bY2;
    bC2 = (bT2 - bSUM2) - bY2;
    bSUM2 = bT2;

    if (evenodd && (k % 2 == 0)) {
      /* NORM2 b EVEN */
      bY2e = belem - bC2e;
      bT2e = bSUM2e + bY2e;
      bC2e = (bT2e - bSUM2e) - bY2e;
      bSUM2e = bT2e;
    }
    
    if (evenodd && (k % 2 == 1)) {
      /* NORM2 b ODD */
      bY2o = belem - bC2o;
      bT2o = bSUM2o + bY2o;
      bC2o = (bT2o - bSUM2o) - bY2o;
      bSUM2o = bT2o;
    }

    /* NORM1 b-i */
    Y1 = sqrtl(x * x + y * y) - C1; /* qui sommo i valori assoluti delle componenti */
    T1 = SUM1 + Y1;
    C1 = (T1 - SUM1) - Y1;
    SUM1 = T1;

    /* NORM1 b */
    bY1 = sqrtl(belem) - bC1;
    bT1 = bSUM1 + bY1;
    bC1 = (bT1 - bSUM1) - bY1;
    bSUM1 = bT1;

    if (evenodd && (k % 2 == 0)) {
      /* NORM1 b EVEN */
      bY1e = sqrtl(belem) - bC1e;
      bT1e = bSUM1e + bY1e;
      bC1e = (bT1e - bSUM1e) - bY1e;
      bSUM1e = bT1e;
    }
    
    if (evenodd && (k % 2 == 1)) {
      /* NORM1 b ODD */
      bY1o = sqrtl(belem) - bC1o;
      bT1o = bSUM1o + bY1o;
      bC1o = (bT1o - bSUM1o) - bY1o;
      bSUM1o = bT1o;
    }

    /* NORM00 b - i */
    DTEMP = sqrtl(x * x + y * y);
    if (DTEMP > DMAX) {
      DMAX = DTEMP;
    }

    /* NORM00 b */
    BTEMP = sqrtl(belem);
    if (BTEMP > BMAX) {
      BMAX = BTEMP;
    }

    if (evenodd && (k % 2 == 0) && (BTEMP > BMAXe)) {
      BMAXe = BTEMP;
    }

    if (evenodd && (k % 2 == 1) && (BTEMP > BMAXo)) {
      BMAXo = BTEMP;
    }

  }

  m->norm2 = sqrtl(SUM2);
  m->bnorm2 = sqrtl(bSUM2);
  if (evenodd) {
    m->bnorm2even = bSUM2e;
    m->bnorm2odd = bSUM2o;
  }

  m->norm1 = SUM1;
  m->bnorm1 = bSUM1;
  if (evenodd) {
    m->bnorm1even = bSUM1e;
    m->bnorm1odd = bSUM1o;
  }
  
  m->norm00 = DMAX;
  m->bnorm00 = BMAX;
  if (evenodd) {
    m->bnorm00even = BMAXe;
    m->bnorm00odd = BMAXo;
  }

  if (fabsl(m->bnorm2) > 0.000001) {
    m->prec2 = m->norm2 / m->bnorm2;
  } else {
    m->prec2 = 0.0;
  }

  if (fabsl(m->bnorm1) > 0.000001) {
    m->prec1 = m->norm1 / m->bnorm1;
  } else {
    m->prec1 = 0.0;
  }

  if (fabsl(m->bnorm00) > 0.000001) {
    m->prec00 = m->norm00 / m->bnorm00;
  } else {
    m->prec00 = 0.0;
  }
  
  unmap_File2Array(b, size, fd_b);
  unmap_File2Array(i, size, fd_i);

  return 1;
}

fftwl_complex * mkAK(fftwl_complex *fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_AK) {

  long double pi = 3.1415926535897932384626433832795028841971693993751;
  long double sinvalue;
  long double cosvalue;
  long double twist = 0.0L;
  long double angle = 0.0L; 
  long double c;
  long double lq = (long double) q;
  ptrdiff_t a;
  ptrdiff_t k;

  /* 
   * STEP: mkAK
   *  need: fftin mmapped array available for writing
   *  provide: input data AK (in fftin) to compute <basename>-AKfft.ld
   */
  
  clock_t startclock = clock();  /* start clock */  
  angle = (long double) pi/nhalf;
  twist = 0.0L; 
  a = 1; 
  for (k = 0; k < nhalf ; k++){
    /*g^k -R g^(k+m) \equiv g^k  -R (q-1)g^k \equiv g^k -R (-g^k) \equiv g^k -R (q -g^k) = 2*g^k -R q */

    /* -- 20210107 -- ERA: c = (long double) (2*a-q); Vedi mail del 20210102 */ 
    // c = (long double) ((2.0 * ((long double)a) / ((long double)q)) - 1);
    c = (long double) (2.0L * (long double) a - lq) / lq;
    // c = (long double) a/q;
    // c = (long double) 2.0L * c - 1.0L;
    cosvalue = cosl(twist);
    sinvalue = - sinl(twist);  /* depends on the sign = -1 of the transform */
    fftin[k][REAL] = c * cosvalue;
    fftin[k][IMAG] = c * sinvalue;
    a = (a * primroot) % q;
    twist += angle;
  }
  clock_t endclock = clock();  /* end clock */
  print_time("AK computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwl_complex * ak = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_AK, 'r', fd_AK);

  return ak;

}

fftwl_complex * fftAK(fftwl_complex *fftin, fftwl_complex *fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_AKfft) {
  
  /* STEP: fftAK 
   *  need: fftin mmapped array with ak start data 
   *  provide: fft in fftout, file <basename>-AKoddfft.ld if basename != NULL
   */

  clock_t startfft = clock();
  fftwl_execute(plan);
  clock_t endfft  = clock();
  print_time("FFT(AK)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /*
  ptrdiff_t q = 2 * nhalf + 1;
  for (ptrdiff_t k = 0; k < nhalf; k++) {
    fftout[k][REAL] *= q;
    fftout[k][IMAG] *= q;
  }
  */
  
  fftwl_complex * akfft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTAK, 'r', fd_AKfft);

  return akfft;

}

struct result oddResult(fftwl_complex *lngammaodd_fft, fftwl_complex *ak_fft, ptrdiff_t nhalf, long int q) {

  struct result myresult;
  
  printf("%s", "Computing the sum and minmax over odd characters"); 
  fflush(stdout);
  
  clock_t startcompodd = clock(); /* start clock */

  long double pi = 3.1415926535897932384626433832795028841971693993751;
  long double oddresult = 0;
  long double res;
  long double lq = (long double) q;
  long double akREq;
  long double akIMq;
  long double Kummer_res;
  long double Kummer;
  long double realpart;
  long double imagpart;
  long double Minimumoddresult;
  long double Maximumoddresult;
  long double MinimumLoddresult;
  long double MaximumLoddresult;
  long double MinimumLprimeoddresult;
  long double MaximumLprimeoddresult;
  long double Minimum = LDBL_MAX;/* odd characters */
  long double Maximum = 0; 
  long double MinimumL = LDBL_MAX;/* odd characters */
  long double MaximumL = 0;
  long double MinimumLprime = LDBL_MAX;/* odd characters */
  long double MaximumLprime = 0; 
  long int jmin = 0;
  long int jmax = 0;
  long int jminL = 0;
  long int jmaxL = 0;
  long int jminLprime = 0;
  long int jmaxLprime = 0;
  long double num     = 0;
  long double num1    = 0;
  long double den     = 0;
  long double quot    = 0;   
  long double aux     = 0; 
  long double frac    = 0;   
  long double frac1   = 0;   
  long double corr    = logl(2*pi)+Eulergamma;  
  // long double corrq   = corr/q; // corr/q;
  long double Kummer_realcorrection = (logl(pi) - (1.5) * logl((long double) q)) * nhalf;

  res = 0.0;
  Kummer_res = 0.0;
  oddresult = 0.0;

  /* computes |(a+ib)/(c+id)|= |a+ib|/|c+id| and the real part of (a+ib)/(c+id); the imaginary part is 0 by def */
  
  for (long int i = 0;  i < nhalf; ++i) {
    // -- codice rimpiazzato LR 20210115
    // num = q * (lngammaodd_fft[i][REAL] * ak_fft[i][REAL] + lngammaodd_fft[i][IMAG] * ak_fft[i][IMAG]);
    // den = q * q * (ak_fft[i][REAL] * ak_fft[i][REAL] + ak_fft[i][IMAG] * ak_fft[i][IMAG]);
    akREq = lq * ak_fft[i][REAL];
    akIMq = lq * ak_fft[i][IMAG];
    num =  (lngammaodd_fft[i][REAL] * akREq + lngammaodd_fft[i][IMAG] * akIMq); 
    den =  (akREq * akREq + akIMq * akIMq);
    // -- fine rimpiazzo
    
    res = res + num / den;
    realpart = corr * ak_fft[i][REAL] + lngammaodd_fft[i][REAL];
    imagpart = corr * ak_fft[i][IMAG] + lngammaodd_fft[i][IMAG];
    aux = realpart * realpart + imagpart * imagpart;
    Kummer_res = Kummer_res + (0.5) * logl(den);
    // -- codice rimpiazzato: LR 20210115
    // num1 = q * (lngammaodd_fft[i][IMAG] * ak_fft[i][REAL] - lngammaodd_fft[i][REAL] * ak_fft[i][IMAG]);
    num1 = (lngammaodd_fft[i][IMAG] * akREq - lngammaodd_fft[i][REAL] * akIMq);
    // -- fine rimpiazzo
    frac = corr + q * num / den;
    frac1 = q * num1 / den;
    quot = frac * frac + frac1 * frac1;
    if (quot < Minimum) { Minimum = quot; jmin = 2*i+1;};
    if (quot > Maximum) { Maximum = quot; jmax = 2*i+1;};
    if (den < MinimumL) { MinimumL = den; jminL = 2*i+1;}; 
    if (den > MaximumL) { MaximumL = den; jmaxL = 2*i+1;};
    if (aux < MinimumLprime) { MinimumLprime = aux; jminLprime = 2*i+1;};
    if (aux > MaximumLprime) { MaximumLprime = aux; jmaxLprime = 2*i+1;};  
  }
  
  Minimumoddresult = sqrtl(Minimum); /* min for odd characters */
  Maximumoddresult = sqrtl(Maximum); /* max for odd characters */
  printf("\n *** indexes for minimal and maximal odd characters L'/L (1,chi) \n ");
  printf("chiminodd index = ");
  printf("%ld\n",jmin);
  printf("chimaxodd index = ");
  printf("%ld\n",jmax);
  
  oddresult = q * res; /* correction factors for odd characters */

  fprintf(stdout, "Kummer_realcorrection = %LE\n", Kummer_realcorrection);  
  fprintf(stdout, "Kummer_res = %LE\n", Kummer_res);
  Kummer = expl(Kummer_res + Kummer_realcorrection);
  fprintf(stdout, "Kummer = %LE\n", Kummer);
  
  long double radq = sqrtl(q);
  MinimumLoddresult = pi * sqrtl(MinimumL)/(q*radq); /* min L for odd characters */
  MaximumLoddresult = pi * sqrtl(MaximumL)/(q*radq); /* max L for odd characters */
  printf("\n *** indexes for minimal and maximal odd characters L (1,chi) \n ");
  printf("chiminoddL index = ");
  printf("%ld\n",jminL);
  printf("chimaxoddL index = ");
  printf("%ld\n",jmaxL);
  
  MinimumLprimeoddresult=pi*sqrtl(MinimumLprime)/radq; /* min L' for odd characters */
  MaximumLprimeoddresult=pi*sqrtl(MaximumLprime)/radq; /* max L' for odd characters */
  printf("\n *** indexes for minimal and maximal odd characters L'(1,chi) \n ");
  printf("chiminoddLprime index = ");
  printf("%ld\n",jminLprime);
  printf("chimaxoddLprime index = ");
  printf("%ld\n",jmaxLprime);
 

  myresult.res = oddresult;
  myresult.max = Maximumoddresult;
  myresult.min = Minimumoddresult;  
  myresult.lambda = 0;
  myresult.Kummer = Kummer;
  myresult.maxL = MaximumLoddresult;
  myresult.minL = MinimumLoddresult;
  myresult.maxLprime = MaximumLprimeoddresult;
  myresult.minLprime = MinimumLprimeoddresult;

  clock_t endcompodd = clock(); /* stop clock */
  printf("\n *** First: oddresult = %32.30Lf, (res = %24.28Lf)\n", oddresult, res);
  print_time("OddResult computation", startcompodd, endcompodd);
  
  fflush(stdout);

  return myresult;
}

fftwl_complex * mkS(fftwl_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, int precision, const char * basename, int * fd_S) {
  
  /* STEP: S 
   *  need: fftin array available for writing
   *  provide: S data in fftin to compute <basename>-Sfft.ld
   *  provide: 
   */
  
  clock_t startcomputeS = clock();
  
  /* inserting the null imaginary parts */
  long int s_a = lpow_mod(primroot, 0, q);
  long double s_aq = 0.0;
  for  (long int k = 0; k < nhalf + 2; k++) {
    s_aq = (long double) ((long double) s_a / (long double) q);
    fftin[k][REAL] = s(s_aq, precision, 0); 
    fftin[k][IMAG] = 0.0L;
    s_a = (s_a * primroot) % q;
  }
  
  clock_t endcomputeS = clock();
  fflush(stdout);
  print_time("S computation", startcomputeS, endcomputeS);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  fftwl_complex * S = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_S, 'r', fd_S);

  return S;
}

fftwl_complex * fftS(fftwl_complex * fftin, fftwl_complex * fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_Sfft) {
  
  /* STEP: fftS 
   *  need: fftin mmapped array with S data 
   *  provide: file <basename>-Sfft.ld (by copying fftout)
   */
 
  clock_t startfft  = clock();
  fftwl_execute(plan);
  clock_t endfft = clock();
  print_time("Sfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwl_complex * S_fft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTS, 'r', fd_Sfft);

  return S_fft;
}

fftwl_complex * mkLGeven(fftwl_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char *basename, int * fd_LGeven) {
  
  /*
   * STEP: mkLGeven 
   *  need: fftin mmapped array
   *  provide: file <basename>-LGeven.ld
   *
   */
  
  long double pi = 3.1415926535897932384626433832795028841971693993751;
  // long double logpi = logl(pi);
  long double d;
  long int a;  
  long int k;

  clock_t startcomp = clock();
  a = 1; 
  for (k = 0; k < nhalf; k++){  
     d = (long double) a/q ;
     fftin[k][REAL]= - logl(sinl(pi*d)); /* logpi is transformed to 0 by orthogonality of Dirichlet characters*/
     /* fftin[k][REAL]= logpi - logl(sinl(pi*d)); */ /* values of log(gamma(a_k/q) via the reflection formula*/
     fftin[k][IMAG] = 0.0L;  
     a = (a * primroot) % q;
  }
  clock_t endcomp = clock();
  print_time("mkLGeven", startcomp, endcomp);
  fflush(stdout);
  
  if (basename == NULL) {
    /* copy of the file not requested, output in fftin */
    return NULL;
  }

  fftwl_complex * LGeven = (fftwl_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwl_complex), basename, N_LGEVEN, 'r', fd_LGeven);

  return LGeven;

}

fftwl_complex * fftLGeven(fftwl_complex *fftin, fftwl_complex *fftout, fftwl_plan plan, long int nhalf, const char *basename, int *fd_LGevenfft) {
  
  /* STEP: fftLGeven 
   *  need: fftin mmapped array with LGeven data
   *  provide: file <basename>-LGevenfft.ld (by renaming fftout)
   */
  
  clock_t startfft = clock();
  fftwl_execute(plan);
  clock_t endfft = clock();
  print_time("LGevenfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwl_complex * LGevenfft = (fftwl_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwl_complex), basename, N_FFTLGEVEN, 'r', fd_LGevenfft);

  return LGevenfft;

}

struct result evenResult(fftwl_complex *lngamma_evenfft, fftwl_complex *S_fft, ptrdiff_t nhalf, long int q) {

  /* STEP: evenResult */
  
  /* Computing EK^+(q) */
  /* Computing min_{chi\ne chi_0} \vert L'/L(1,\chi) \vert =: M(q) */
  /* sum starts from i=1 to avoid the trivial character contribution */
  /* values_trasf contains the DFT of lngamma_even; due to the simmetries the output has half elements than the input */
  /* values_trasf[i][REAL]=values_trasf[nhalf-i][REAL];  values_trasf[i][IMAG] = -values_trasf[nhalf-i][IMAG] */
  /* S_trasf contains the DFT of S; due to the simmetries the output has half elements than the input */
  /* S_trasf[i][REAL]=S_trasf[nhalf-i][REAL];  S_lngamma_even_tras[i][IMAG] = -S_trasf[nhalf-i][IMAG] */
  /*   SO THEY HAVE THE SAME ABSOLUTE VALUES  */

  struct result myresult;
  
  int flagq = 0;
  long double lambdares    = 0;
  long double lambdaresult = 0;
  long double evenresult = 0;
  long double res = 0;
  long double realpart = 0;
  long double imagpart = 0;
  long double Minimumevenresult = 0;
  long double Maximumevenresult = 0;
  long double Minimum = LDBL_MAX; /* even characters */
  long double Maximum = 0; 
  long double MinimumLevenresult = 0;
  long double MaximumLevenresult = 0;
  long double MinimumL = LDBL_MAX; /* even characters */
  long double MaximumL = 0; 
  long double MinimumLprimeevenresult = 0;
  long double MaximumLprimeevenresult = 0;
  long double MinimumLprime = LDBL_MAX; /* even characters */
  long double MaximumLprime = 0; 
  long int jmin = 0;
  long int jmax = 0;
  long int jminL = 0;
  long int jmaxL = 0;
  long int jminLprime = 0;
  long int jmaxLprime = 0;
  long double num     = 0;
  long double num1    = 0;
  long double den     = 0;
  long double quot    = 0;   
  long double aux    = 0;   
  long double frac    = 0;   
  long double frac1   = 0;   
  long double corr    = logl(2*pi)+Eulergamma;
  long double twicecorr    = 2 * corr;

  // ptrdiff_t nquarter = nhalf / 2;
  // ptrdiff_t m;

  if (nhalf % 2 == 0) {flagq=1;};
  res = 0;
  evenresult = 0;

  printf("Computing the sum and minmax over even characters ..."); 
  clock_t startcompeven = clock();

  /* sum starts from i=1 to avoid the trivial character contribution */ 
     
  for (long int i=1;  i < nhalf; ++i) { 
      num = S_fft[i][REAL] * lngamma_evenfft[i][REAL] +  S_fft[i][IMAG]*lngamma_evenfft[i][IMAG];
      den = lngamma_evenfft[i][REAL] *lngamma_evenfft[i][REAL] + lngamma_evenfft[i][IMAG]*lngamma_evenfft[i][IMAG];
      res = res + num / den;
      num1 = S_fft[i][IMAG] * lngamma_evenfft[i][REAL] - S_fft[i][REAL]*lngamma_evenfft[i][IMAG];
      realpart = twicecorr * lngamma_evenfft[i][REAL] + S_fft[i][REAL];
      imagpart = twicecorr * lngamma_evenfft[i][IMAG] + S_fft[i][IMAG];
      aux = realpart * realpart + imagpart * imagpart;
      frac = corr + num/(2*den);
      frac1 = num1 / (2*den);
      quot = frac * frac + frac1 * frac1; 
      if (quot < Minimum) { Minimum = quot; jmin = 2*i;};
      if (quot > Maximum) { Maximum = quot; jmax = 2*i;};
      if (den < MinimumL) { MinimumL = den; jminL = 2*i;}; 
      if (den > MaximumL) { MaximumL = den; jmaxL = 2*i;};
      if (aux < MinimumLprime) { MinimumLprime = aux; jminLprime = 2*i;};
	  if (aux > MaximumLprime) { MaximumLprime = aux; jmaxLprime = 2*i;};  
      if ((flagq == 1) & (i%2 == 0)) {lambdares = lambdares + num/den; };  /* compute lambda_q on some special characters */

  }   
  
  Minimumevenresult = sqrtl(Minimum); /* min for even characters */
  Maximumevenresult = sqrtl(Maximum); /* max for even characters */
  printf("\n *** indexes for minimal and maximal even characters L'/L (1,chi) \n ");
  printf("chimineven index = ");
  printf("%ld\n",jmin);
  printf("chimaxeven index = ");
  printf("%ld\n",jmax);
  
  long double radq= sqrtl(q);
  MinimumLevenresult = 2 * sqrtl(MinimumL) / radq; /* min L for even characters */
  MaximumLevenresult = 2 * sqrtl(MaximumL) / radq; /* max L for even characters */
  printf("\n *** indexes for minimal and maximal even characters L(1,chi) \n ");
  printf("chiminevenL index = ");
  printf("%ld\n",jminL);
  printf("chimaxevenL index = ");
  printf("%ld\n",jmaxL);
  
  MinimumLprimeevenresult = sqrtl(MinimumLprime) / radq; /* min L' for even characters */
  MaximumLprimeevenresult = sqrtl(MaximumLprime) / radq; /* max L' for even characters */
  printf("\n *** indexes for minimal and maximal even characters L'(1,chi) \n ");
  printf("chiminevenLprime index = ");
  printf("%ld\n",jminLprime);
  printf("chimaxevenLprime index = ");
  printf("%ld\n",jmaxLprime);
  
  evenresult = res/2.0;
  if (flagq == 1) {lambdaresult = lambdares/2.0; }; 

  myresult.res = evenresult;
  myresult.max = Maximumevenresult;
  myresult.min = Minimumevenresult;
  myresult.lambda = 0;
  if (flagq == 1) {myresult.lambda = lambdaresult; };
  myresult.Kummer = 0;
  myresult.maxL = MaximumLevenresult;
  myresult.minL = MinimumLevenresult;
  myresult.maxLprime = MaximumLprimeevenresult;
  myresult.minLprime = MinimumLprimeevenresult;

  clock_t endcompeven = clock();

  printf("\n *** Second: evenresult = %32.30Lf, (res = %24.28Lf)\n", evenresult, res);
  print_time("EvenResult computation", startcompeven, endcompeven);
  
  fflush(stdout);

  return myresult;
}

int printMeasure(struct measure m, const char *name, FILE *stream, int evenodd) {
  fprintf(stream, "Measures for %s\n", name);

  fprintf(stream, " NDIFF_2 = Norm2(%s - fft(fft(%s)) = %32.30LE\n", name, name, m.norm2);
  fprintf(stream, " NBASE_2 = Norm2(%s) = %32.30LE\n", name, m.bnorm2);
  fprintf(stream, " Inverse precision for %s in Norm2: NDIFF_2 / NBASE_2 = %32.30LE\n", name, m.prec2);
  
  fprintf(stream, " NDIFF_1 = Norm1(%s - fft(fft(%s)) = %32.30LE\n", name, name, m.norm1);
  fprintf(stream, " NBASE_1 = Norm1(%s) = %32.30LE\n", name, m.bnorm1);
  fprintf(stream, " Inverse precision for %s in Norm1: NDIFF_1 / NBASE_1 = %32.30LE\n", name, m.prec1);
  
  fprintf(stream, " NDIFF_0 = Norm00(%s - fft(fft(%s)) = %32.30LE\n", name, name, m.norm00);
  fprintf(stream, " NBASE_0 = Norm00(%s) = %32.30LE\n", name, m.bnorm00);
  fprintf(stream, " Inverse precision for %s in Norm0: NDIFF_0 / NBASE_0 = %32.30LE\n", name, m.prec00);

  if (evenodd) {
    fprintf(stream, " NBASE_2_even = Norm2(%s_even)  = %32.30LE\n", name, m.bnorm2even);
    fprintf(stream, " NBASE_2_odd  = Norm2(%s_odd)   = %32.30LE\n", name, m.bnorm2odd);
    fprintf(stream, " NBASE_1_even = Norm1(%s_even)  = %32.30LE\n", name, m.bnorm1even);
    fprintf(stream, " NBASE_1_odd  = Norm1(%s_odd)   = %32.30LE\n", name, m.bnorm1even);
    fprintf(stream, " NBASE_0_even = Norm00(%s_odd)  = %32.30LE\n", name, m.bnorm00even);
    fprintf(stream, " NBASE_0_odd  = Norm00(%s_even) = %32.30LE\n", name, m.bnorm00odd);
  }

  
  fprintf(stream, "\n");
  
  return 1;
}


int printNorms(struct norms norm) {
  /* print norm result */
  printMeasure(norm.norm_s, "S", stdout, NO);
  printMeasure(norm.norm_ak, "AK", stdout, NO);
  printMeasure(norm.norm_lgodd, "LGodd", stdout, NO);
  printMeasure(norm.norm_lgeven, "LGeven", stdout, NO);
  return 1;
}

int printResultsORIG(long int q, ptrdiff_t nhalf, struct result oddresult, struct result evenresult) {

  /** RESULTS **/
  printf("*** RESULTS:\n\n");
  fflush(stdout);
  long double EK = Eulergamma + evenresult.res + oddresult.res + (Eulergamma + logl(2*pi)) *(q-2);
  long double EK_plus = nhalf*Eulergamma + (nhalf-1)* logl(2*pi) + evenresult.res;
  long double EK_diff = oddresult.res + (Eulergamma + log(2*pi)) * nhalf;
  
  if (EK < 0 || EK_plus < 0 ) {
    FILE * negativeEKfile;
    negativeEKfile = fopen ("../negativeEK.txt","a");
    fprintf (negativeEKfile, "q = %td, ", q);
    fprintf (negativeEKfile, "EK = %32.30Lf, ", EK); 
    fprintf (negativeEKfile, "EK+ = %32.30Lf, ", EK_plus); 
    fprintf (negativeEKfile, "EK-EK+ = %32.30Lf\n", EK_diff); 
    fclose (negativeEKfile);
  };
  
  /* print EK_q */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", EK);
  
  /* print EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")^+ = ");
  printf("%32.30Lf\n", EK_plus); 
  
  /* print EK_q- EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ") - ");
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")^+ = ");
  printf("%32.30Lf\n", EK_diff); 

  printf("%s\n", "--------------");
  
  /* print min_q */
  printf("%s", "min_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.min); 

  /* print min_q */
  printf("%s", "min_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.min);  

  /* print min_q */
  printf("%s", "min_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fminl(evenresult.min, oddresult.min));    
  
  return 1;
}

int printResults(long int q, ptrdiff_t nhalf, struct result oddresult, struct result evenresult) {
  
  int flagq = 0;
  
  if ((q-1) % 4 == 0) {
    flagq = 1;
  }; /*flagq = 1 means q = 1 mod 4 */
  
  /** RESULTS **/
  printf("*** RESULTS:\n\n");
  fflush(stdout);
  long double EK = Eulergamma + evenresult.res + oddresult.res + (Eulergamma + logl(2*pi)) *(q-2);
  long double EK_plus = nhalf*Eulergamma + (nhalf-1)* logl(2*pi) + evenresult.res;
  long double EK_diff = oddresult.res + (Eulergamma + logl(2*pi)) * nhalf;
  long double lambda = 0;
  long double lambda_diff = 0;
  if (flagq == 1) {
      lambda = (1 + floor((q-3)/4)) * Eulergamma + floor((q-3)/4) * logl(2*pi)+ evenresult.lambda;
      lambda_diff = EK_plus - lambda;
  };
  
  if (EK < 0 || EK_plus < 0 ) {
    FILE * negativeEKfile;
    negativeEKfile = fopen ("../negativeEK.txt","a");
    fprintf (negativeEKfile, "q = %td, ", q);
    fprintf (negativeEKfile, "EK = %32.30Lf, ", EK); 
    fprintf (negativeEKfile, "EK+ = %32.30Lf, ", EK_plus); 
    fprintf (negativeEKfile, "EK-EK+ = %32.30Lf", EK_diff); 
    if (flagq == 1) {
        fprintf (negativeEKfile, ", lambda = %32.30Lf", lambda); 
        fprintf (negativeEKfile, ", EK+ - lambda = %32.30Lf", lambda_diff);
    };
    fprintf (negativeEKfile, "\n");
    fclose (negativeEKfile);
  };

  printf("%s\n", "--------------");
  /* print EK_q */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", EK);
  
  /* print EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")plus = ");
  printf("%32.30Lf\n", EK_plus); 
  
  /* print EK_q- EK_q^+ */
  printf("%s", "EK(");
  printf("%td",q);
  printf("%s", ")diff ="); 
  printf("%32.30Lf\n", EK_diff); 

  printf("%s\n", "--------------");
  
  if (flagq == 1) {
    /* print lambda_q */
    printf("%s", "lambda(");
    printf("%td",q);
    printf("%s", ") = "); 
    printf("%32.30Lf\n", lambda);
     
    
    /* print EK_q^+ - lambdqa_q */
    printf("%s", "EK(");
    printf("%td",q);
    printf("%s", ")^+ - "); 
    printf("%s", "lambda(");
    printf("%td",q);
    printf("%s", ") = "); 
    printf("%32.30Lf\n",  lambda_diff);
    
    printf("%s\n", "--------------");
  };
  
  /* print min_q */
  printf("%s", "min_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.min); 

  /* print min_q */
  printf("%s", "min_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.min);  

  
  /* print min_q */
  printf("%s", "min_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fminl(evenresult.min, oddresult.min));    
  
  printf("%s\n", "--------------");
  
  /* print max_q */
  printf("%s", "max_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.max); 

  /* print max_q */
  printf("%s", "max_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.max);  

  
  /* print min_q */
  printf("%s", "max_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fmaxl(evenresult.max, oddresult.max));   
  
  printf("%s\n", "--------------");  
  
  printf("%s", "Kummer_ratio(");
  printf("%td",q);
  printf("%s", ") = ");
  /*printf("%Lf\n", Kummer);*/
  printf("%32.30Lf\n", oddresult.Kummer);  
  
  printf("%s\n", "--------------");
  
    /* print minL_q */
  printf("%s", "minL_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.minL); 

  /* print minL_q */
  printf("%s", "minL_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.minL);  

  
  /* print minL_q */
  printf("%s", "minL_{chi neq chi_0}|L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fminl(evenresult.minL, oddresult.minL));    
  
  printf("%s\n", "--------------");
  
  /* print maxL_q */
  printf("%s", "maxL_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.maxL); 

  /* print maxL_q */
  printf("%s", "maxL_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.maxL);  

  
  /* print maxL_q */
  printf("%s", "maxL_{chi neq chi_0}|L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fmaxl(evenresult.maxL, oddresult.maxL));   
  
  printf("%s\n", "--------------");   
  
     /* print minLprime_q */
  printf("%s", "minLprime_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.minLprime); 

  /* print minLprime_q */
  printf("%s", "minLprime_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.minLprime);  

  
  /* print minLprime_q */
  printf("%s", "minLprime_{chi neq chi_0}|Lprime(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fminl(evenresult.minLprime, oddresult.minLprime));    
  
  printf("%s\n", "--------------");
  
  /* print maxLprime_q */
  printf("%s", "maxLprime_even(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", evenresult.maxLprime); 

  /* print maxLprime_q */
  printf("%s", "maxLprime_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", oddresult.maxLprime);  

  
  /* print maxLprime_q */
  printf("%s", "maxLprime_{chi neq chi_0}|L'(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  printf("%32.30Lf\n", fmaxl(evenresult.maxLprime, oddresult.maxLprime));   
  
  printf("%s\n", "--------------");   
  
  return 1;
}

void apriori(long int q, int precision, FILE * stream) {
  long double N1LGeven = 0.0;
  long double N00LGeven = 0.0;
  /* long double N1LGodd = 0.0; */
  long double N00LGodd = 0.0; /**/
  long double Z0 = -2.00635645590858485121010002;
  long double N1Seven = 0.0;
  long double N00Seven = 0.0;
  long double logq = logl((long double) q);
  long double lq = (long double) q;
  long double N1AKodd = 0.0L;
  long double N00AKodd = 0.0L;
  

  
  N1LGeven = 0.5 * ((q - 1) * logl(2.0) - logq);
  N00LGeven = - logl(sinl(pi / lq));
  N1Seven = - Z0 * (q - 1) - logq * logl(2.0 * pi) - 0.5 * (logq * logq);
  N00Seven = - 2.0 * s(1.0 / lq, precision, 0) + s(1.0 - 1.0/lq, precision, 0);
  N1AKodd = (q - 1) * 0.25 * (q - 1) / lq;
  N00AKodd = (q - 2) / lq;
  /* N1LGodd = -1; */
  N00LGodd = 2.0 * lgammal(1.0 / lq) + logl(sin(pi / lq) / pi);

  fprintf(stream, "*** A priori (closed formulas) norms ***\n");
  
  fprintf(stream, "N1LGeven  = %32.30LE\n", N1LGeven);
  fprintf(stream, "N00LGeven = %32.30LE\n", N00LGeven);

  // fprintf(stream, "N1LGodd   = %32.30LE\n", N1LGodd);
  fprintf(stream, "N00LGodd  = %32.30LE\n", N00LGodd);
  
  fprintf(stream, "N1Seven   = %32.30LE\n", N1Seven);
  fprintf(stream, "N00Seven  = %32.30LE\n", N00Seven);
  fprintf(stream, "N1AKodd   = %32.30LE\n", N1AKodd);
  fprintf(stream, "N00AKodd  = %32.30LE\n", N00AKodd);

}


int dostep(const char * step, const char * steps) {
  if (strcasestr(steps, step) != NULL) {
    return 1;
  }
  return 0;
}

int runComputation(long int q, long int primroot, int precision, const char * basename, const char * steps) {

  ptrdiff_t NUM_POINTS = q-1;
  ptrdiff_t nhalf = NUM_POINTS / 2;

  int fd_fftin;
  int fd_fftout;
  int fd_fftinverse;
  int fd_LGodd;
  int fd_LGoddfft;
  int fd_AK;
  int fd_AKfft;
  int fd_S;
  int fd_Sfft;
  int fd_LGeven;
  int fd_LGevenfft;
  
  fftwl_complex * fftin  = NULL; // mkfftin(nhalf, basename, &fd_fftin);
  fftwl_complex * fftout = NULL; // mkfftout(nhalf, basename, &fd_fftout);
  fftwl_complex * fftinverse = NULL; // mkfftinverse(nhalf, basename, &fd_fftinverse);
  fftwl_complex * lgodd = NULL;
  fftwl_complex * lgoddfft = NULL;
  fftwl_complex * ak = NULL;
  fftwl_complex * akfft = NULL;
  fftwl_complex * S = NULL;
  fftwl_complex * Sfft = NULL;
  fftwl_complex * lgeven = NULL;
  fftwl_complex * lgevenfft = NULL;
  
  fftwl_plan plan = NULL; 
  fftwl_plan inverse_plan = NULL; 

  struct result oddresult;
  struct result evenresult;
  struct norms allnorms;
  
  int doall = 0; /* make every step */
  int docompute = 0; /* make compute step */
  
  fprintf(stdout, "runComputation: steps: %s\n", steps);
  
  if (dostep(ALL, steps)) {
    doall = 1;
  }

  if (dostep(COMPUTE, steps)) {
    docompute = 1;
  }
  
  fftin  = mkfftin(nhalf, basename, &fd_fftin);
  fftout = mkfftout(nhalf, basename, &fd_fftout);
  // fftinverse = mkfftinverse(nhalf, basename, &fd_fftinverse);
  
  if (fftin == NULL || fftout == NULL)
    return 0;
  
  if (doall || dostep(COMPUTE, steps)) {
  
    plan = mkplan(fftin, fftout, nhalf);
    if (plan == NULL)
      return 0;

  }

  if (doall || docompute || dostep(ODDRESULT, steps)) {

    fprintf(stdout, "COMPUTE ODD\n");
    
    if (doall || docompute || dostep(MKLGODD, steps)) {
      lgodd     = mkLGodd(fftin, nhalf, q, primroot, basename, &fd_LGodd);       /* qui non servirebbe salvare */
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_LGODD);
      lgodd     = mapfile(N_LGODD, basename, nhalf, &fd_LGodd);
    }
    
    if (doall || docompute || dostep(FFTLGODD, steps)) {
      lgoddfft  = fftLGodd(fftin, fftout, plan, nhalf, basename, &fd_LGoddfft);
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_FFTLGODD);
      lgoddfft  = mapfile(N_FFTLGODD, basename, nhalf, &fd_LGoddfft);
    }
    
    if (doall || docompute || dostep(MKAK, steps)) {
      ak        = mkAK(fftin, nhalf, q, primroot, basename, &fd_AK);             /* qui non servirebbe salvare */
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_AK);
      ak        = mapfile(N_AK, basename, nhalf, &fd_AK);
    }
    
    if (doall || docompute || dostep(FFTAK, steps)) {
      akfft     = fftAK(fftin, fftout, plan, nhalf, basename, &fd_AKfft);        /* qui non servirebbe salvare */
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_FFTAK);
      akfft     = mapfile(N_FFTAK, basename, nhalf, &fd_AKfft);
    }
    
    oddresult   = oddResult(lgoddfft, akfft, nhalf, q);

    writeResult(oddresult, basename, N_ODDRESULT);
    
    unmap_File2Array(lgodd, nhalf, fd_LGodd);
    unmap_File2Array(lgoddfft, nhalf, fd_LGoddfft);
    unmap_File2Array(ak, nhalf, fd_AK);
    unmap_File2Array(akfft, nhalf, fd_AKfft);

  } else {

    oddresult = readResult(basename, N_ODDRESULT);

  }
  
  if (doall || docompute || dostep(EVENRESULT, steps)) {
  
    fprintf(stdout, "COMPUTE EVEN\n");

    if (doall || docompute || dostep(MKS, steps)) {
      S         = mkS(fftin, nhalf, q, primroot, precision, basename, &fd_S);    /* qui non servirebbe salvare */
    } else {
      S         = mapfile(N_S, basename, nhalf, &fd_S);
    }
    
    if (doall || docompute || dostep(FFTS, steps)) {
      Sfft      = fftS(fftin, fftout, plan, nhalf, basename, &fd_Sfft);
    } else {
      Sfft      = mapfile(N_FFTS, basename, nhalf, &fd_Sfft);
    }
    
    if (doall || docompute || dostep(MKLGEVEN, steps)) {
      lgeven    = mkLGeven(fftin, nhalf, q, primroot, basename, &fd_LGeven);      /* qui non servirebbe salvare */
    } else {
      lgeven    = mapfile(N_LGEVEN, basename, nhalf, &fd_LGeven);
    }
    
    if (doall || docompute || dostep(FFTLGEVEN, steps)) {
      lgevenfft = fftLGeven(fftin, fftout, plan, nhalf, basename, &fd_LGevenfft); /* qui non servirebbe salvare */
    } else {
      lgevenfft = mapfile(N_FFTLGEVEN, basename, nhalf, &fd_LGevenfft);
    }
    
    evenresult  = evenResult(lgevenfft, Sfft, nhalf, q);

    writeResult(evenresult, basename, N_EVENRESULT);
    
    unmap_File2Array(S, nhalf, fd_S);
    unmap_File2Array(Sfft, nhalf, fd_Sfft);
    unmap_File2Array(lgeven, nhalf, fd_LGeven);
    unmap_File2Array(lgevenfft, nhalf, fd_LGevenfft);
    
  } else {

    evenresult = readResult(basename, N_EVENRESULT);

  }

  /* Destroy the forward plan */
  if (plan != NULL)
    fftwl_destroy_plan(plan);

  if (fftin != NULL) 
    unmap_File2Array(fftin, nhalf, fd_fftin);

  if (fftout != NULL)
    unmap_File2Array(fftout, nhalf, fd_fftout);

  
  /*  From here if all the computation are availables
   *  we compute the inverse of the FFTs and the
   *  the norms of the differences: V - FFT(FFT(V))
   *  for the various input vectors. All mapped
   *  file should be closed
   */

  if (doall
      || dostep(INV, steps)
      || dostep(ILGODD, steps) 
      || dostep(IAK, steps) 
      || dostep(IS, steps) 
      || dostep(ILGEVEN, steps)) {

    fftin  = mkfftin(nhalf, basename, &fd_fftin);
    fftout = mkfftout(nhalf, basename, &fd_fftout);
    // fftinverse = mkfftinverse(nhalf, basename, &fd_fftinverse);
    if (fftin == NULL || fftout == NULL)
      return 0;
 
    inverse_plan = mkinverseplan(fftin, fftout, nhalf);
    if (inverse_plan == NULL)
      return 0;

  }

  if (doall || dostep(INV, steps) || dostep(ILGODD, steps)) {
    fprintf(stdout, "INVERSE LGodd\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTLGODD,  N_FFTLGODD_INV,  &fdinverse);
  }
  
  if (doall || dostep(INV, steps) || dostep(IAK, steps)) {
    fprintf(stdout, "INVERSE AK\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTAK,     N_FFTAK_INV,     &fdinverse);
  }
  
  if (doall || dostep(INV, steps) || dostep(IS, steps)) {
    fprintf(stdout, "INVERSE S\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTS,      N_FFTS_INV,      &fdinverse);
  }

  if (doall || dostep(INV, steps) || dostep(ILGEVEN, steps)) {
    fprintf(stdout, "INVERSE LGeven\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTLGEVEN, N_FFTLGEVEN_INV, &fdinverse);
  }

  if (inverse_plan == NULL) {
    fftwl_destroy_plan(inverse_plan);
  }

  if (fftin != NULL) 
    unmap_File2Array(fftin, nhalf, fd_fftin);

  if (fftout != NULL)
    unmap_File2Array(fftout, nhalf, fd_fftout);

  fftwl_cleanup();

  if (doall || dostep(NORMS, steps) || dostep(SNORM, steps)) {
    fprintf(stdout, "\nNORM OF S ... \n");
    computenormskahan(basename, N_S,      N_FFTS_INV,      nhalf, &(allnorms.norm_s), YES);
    printMeasure(allnorms.norm_s, "S", stdout, NO);
  }
    
  if (doall || dostep(NORMS, steps) || dostep(AKNORM, steps)) {
    fprintf(stdout, "\nNORM OF AK ... \n");
    computenormskahan(basename, N_AK,     N_FFTAK_INV,     nhalf, &(allnorms.norm_ak), YES);
    printMeasure(allnorms.norm_ak, "AK", stdout, NO);
  }

  if (doall || dostep(NORMS, steps) || dostep(LGODDNORM, steps)) {
    fprintf(stdout, "\nNORM OF LGODD ... \n");
    computenormskahan(basename, N_LGODD,  N_FFTLGODD_INV,  nhalf, &(allnorms.norm_lgodd), NO);
    printMeasure(allnorms.norm_lgodd, "LGodd", stdout, NO);
  }
  
  if (doall || dostep(NORMS, steps) || dostep(LGEVENNORM, steps)) {
    fprintf(stdout, "\nNORM OF LGEVEN ... \n");
    computenormskahan(basename, N_LGEVEN, N_FFTLGEVEN_INV, nhalf, &(allnorms.norm_lgeven), NO);
    printMeasure(allnorms.norm_lgeven, "LGeven", stdout, NO);
  }

  printResults(q, nhalf, oddresult, evenresult);

  apriori(q, precision, stdout);

  return 0;
}

